#ifndef _MEMORY_HPP__
#define _MEMORY_HPP__

namespace std {

template<class T>
class cow_shared_ptr {
	struct cow_shared_ptr_data {
		T * m_Data;
		int m_UseCount;
		cow_shared_ptr_data ( T * data ) : m_Data ( data ), m_UseCount( 0 ) { }
		~cow_shared_ptr_data ( ) {
			delete m_Data;
		}
	};

public:
	cow_shared_ptr ( ) {
		attach ( NULL );
	}

	cow_shared_ptr ( T * data ) {
		attach ( new cow_shared_ptr_data( data ) );
	}

	cow_shared_ptr ( const cow_shared_ptr<T> & other ) {
		attach ( other.m_Data );
	}

	cow_shared_ptr ( cow_shared_ptr<T> && other ) noexcept {
		attach ( other.m_Data );
	}

	~cow_shared_ptr ( ) noexcept {
		deattach ( );
	}

	cow_shared_ptr<T> & operator = ( const cow_shared_ptr<T> & other ) {
		if(this == &other) return *this;

		deattach ( );
		attach ( other.m_Data );

		return *this;
	}

	cow_shared_ptr<T> & operator = ( cow_shared_ptr<T> && other ) noexcept {
		swap(*this, other);
		return *this;
	}

	T * operator -> ( ) {
		make_unique( );
		return m_Data->m_Data;
	}

	T const * operator -> ( ) const {
		return m_Data->m_Data;
	}

	T & operator * ( ) {
		make_unique( );
		return *(m_Data->m_Data);
	}

	T const & operator * ( ) const {
		return *(m_Data->m_Data);
	}

	T * get () {
		make_unique( );
		return m_Data->m_Data;
	}

	T const * get () const {
		return m_Data->m_Data;
	}

	bool unique ( ) const {
		return m_Data == NULL || m_Data->m_UseCount == 1; 
	}

private:
	void attach ( cow_shared_ptr<T>::cow_shared_ptr_data * data ) {
		m_Data = data;
		if(m_Data) m_Data->m_UseCount++;
	}

	void deattach ( ) {
		if(m_Data && --(m_Data->m_UseCount) <= 0) delete m_Data;
		m_Data = NULL;
	}

	void make_unique( ) {
		if(unique( )) return;

		typename cow_shared_ptr<T>::cow_shared_ptr_data* tmp = m_Data;
		deattach( );
		attach(new cow_shared_ptr_data(new T(*(tmp->m_Data))));
	}

	cow_shared_ptr_data * m_Data;

	friend void swap ( cow_shared_ptr<T> & first, cow_shared_ptr<T> & second ) {
		typename cow_shared_ptr<T>::cow_shared_ptr_data* tmp = first.m_Data;
		first.m_Data = second.m_Data;
		second.m_Data = tmp;
	}
};

template<class T>
const cow_shared_ptr<T>& make_const(cow_shared_ptr<T>& ptr) {
	return static_cast<const cow_shared_ptr<T>&>(ptr);
}

template<class T>
const cow_shared_ptr<T>& make_const(const cow_shared_ptr<T>& ptr) {
	return ptr;
}

} /* namespace std */

#endif // _MEMORY_HPP__
